#!/usr/bin/env python3

import argparse
import numpy as np
import pmpred as pm
import time


def main():
    parser = argparse.ArgumentParser(
        description="Using Precision Matrix, Snplists and Sumstats to give a joint effective size"
    )
    parser.add_argument("--pm", type=str, help="Precision Matrix Folder Path")
    parser.add_argument("--snp", type=str, help="Snplists Folder Path")
    parser.add_argument("-s", "--sumstats", type=str, help="Sumstats File")
    parser.add_argument(
        "--rsidname",
        type=str,
        default="rsid",
        help="The header name of rsid. The default is rsid.",
    )
    parser.add_argument(
        "--REFname",
        type=str,
        default="REF",
        help="The header name of REF. The default is REF",
    )
    parser.add_argument(
        "--ALTname",
        type=str,
        default="ALT",
        help="The header name of ALT. The default is ALT",
    )
    parser.add_argument(
        "--betaname",
        type=str,
        default="beta",
        help="The header name of beta. The default is beta",
    )
    parser.add_argument(
        "--sename",
        type=str,
        default="beta_se",
        help="The header name of beta se. The default is beta_se",
    )
    parser.add_argument(
        "--Nname",
        type=str,
        default="N",
        help="The header name of N. The default is N",
    )
    parser.add_argument(
        "--pname",
        type=str,
        default="p",
        help="The header name of p value. The default is p",
    )
    parser.add_argument(
        "--zname",
        type=str,
        default="z",
        help="The header name of z scores. The default is z",
    )
    parser.add_argument(
        "--split",
        type=str,
        default="\t",
        help="The split of sumstats. The default is tap",
    )
    parser.add_argument("-o", "--out", type=str, help="Output File")
    parser.add_argument("--N", type=int, help="Specify N in GWAS sumstats if need")
    parser.add_argument(
        "--burnin",
        type=int,
        default=50,
        help="The number of burn-in iterations used by the Gibbs sampler The default is 50.",
    )
    parser.add_argument(
        "--numiter",
        type=int,
        default=100,
        help="The number of iterations used by the Gibbs sampler. The default is 100.",
    )
    parser.add_argument(
        "--taylor",
        type=float,
        default=0.001,
        help="The number of approximation in taylor expansion. The default is 1e-3.",
    )
    parser.add_argument(
        "--h2",
        type=float,
        default=np.random.rand(),
        help="The genome-wide heritability assumed by PMpred. The default is random",
    )
    parser.add_argument(
        "--p",
        type=float,
        default=np.random.rand(),
        help="The prior probability of non-sparse. The default is random",
    )
    parser.add_argument(
        "--njobs",
        type=int,
        default=-1,
        help="The jobs parallelized. The default is -1",
    )
    parser.add_argument(
        "--method",
        type=str,
        default="pmldpred_auto",
        help="The method we use in PMpred, contain {pmldpred_auto, pmldpred_grid, pmprscs, normalizepm}. The default is pmldpred_auto",
    )
    parser.add_argument(
        "--unnormal",
        action="store_true",
        help="If select, then will not do normalization step for Precision Matrix",
    )
    parser.add_argument(
        "--usepvalue",
        action="store_true",
        help="If select, then will use p value and z scores to generate beta and beta_se",
    )
    parser.add_argument(
        "--rtol",
        type=float,
        default=1e-10,
        help="The precision of gmres algorithm to solve the linear equation. The default is 1e-10",
    )
    parser.add_argument(
        "--prop",
        type=float,
        default=0.1,
        help="The proprotion of the update of h2. The default is 0.1",
    )
    args = parser.parse_args()
    para = pm.generate.get_para()
    para["h2"] = args.h2
    para["p"] = args.p
    para["burn_in"] = args.burnin
    para["num_iter"] = args.numiter
    para["taylor_rtol"] = args.taylor
    para["n_jobs"] = args.njobs
    para["rtol"] = args.rtol
    para["prop"] = args.prop
    head_name = {
        "rsid": args.rsidname,
        "beta": args.betaname,
        "beta_se": args.sename,
        "REF": args.REFname,
        "ALT": args.ALTname,
        "p": args.pname,
        "z": args.zname,
    }
    if args.method == "pmldpred_auto":
        if not args.pm:
            parser.error("You must specify a Precision Matrix Folder Path.")
        if not args.snp:
            parser.error("You must specify a Snplists Folder Path.")
        if not args.sumstats:
            parser.error("You must specify a Sumstats File.")
        start_time = time.time()
        sumstats_list = pm.read.sumstats_read(args.sumstats, args.split, head_name)
        if args.N:
            pm.generate.generate_N_in_sumstats_list(sumstats_list, args.N)
        pm.filter.filter_sumstats(sumstats_list)
        pm.filter.filter_by_unique_sumstats(sumstats_list)
        PM = pm.read.PM_read(args.pm)
        snplist = pm.read.snplist_read(args.snp)
        pm.filter.filter_by_PM(PM, snplist)
        pm.filter.filter_by_unique_snplist(snplist)
        if not args.unnormal:
            pm.filter.normalize_PM_parallel(PM, para)
        sumstats = pm.filter.filter_by_sumstats_parallel(
            PM, snplist, sumstats_list, para
        )
        if args.usepvalue:
            pm.generate.generate_beta_and_se_from_p_and_z(sumstats)
        pm.check.check_same_rsid(snplist, sumstats)
        beta_pmldpred_auto, outpara = pm.pmldpred.pmldpred_auto(
            PM, snplist, sumstats, para
        )
        end_time = time.time()
        pm.write.sumstats_beta_write(
            sumstats,
            beta_pmldpred_auto,
            args.out,
            end_time - start_time,
            outpara,
            head_name,
        )
    elif args.method == "pmldpred_grid":
        if not args.pm:
            parser.error("You must specify a Precision Matrix Folder Path.")
        if not args.snp:
            parser.error("You must specify a Snplists Folder Path.")
        if not args.sumstats:
            parser.error("You must specify a Sumstats File.")
        start_time = time.time()
        sumstats_list = pm.read.sumstats_read(args.sumstats)
        pm.filter.filter_sumstats(sumstats_list)
        pm.filter.filter_by_unique_sumstats(sumstats_list)
        PM = pm.read.PM_read(args.pm)
        snplist = pm.read.snplist_read(args.snp)
        pm.filter.filter_by_PM(PM, snplist)
        pm.filter.filter_by_unique_snplist(snplist)
        if not args.unnormal:
            pm.filter.normalize_PM_parallel(PM, para)
        sumstats = pm.filter.filter_by_sumstats_parallel(
            PM, snplist, sumstats_list, para
        )
        if args.N != None:
            pm.generate.generate_N(sumstats, args.N)
        pm.check.check_same_rsid(snplist, sumstats)
        beta_pmldpred_grid, outpara = pm.pmldpred.pmldpred_grid(
            PM, snplist, sumstats, para
        )
        end_time = time.time()
        pm.write.sumstats_beta_write(
            sumstats, beta_pmldpred_grid, args.out, end_time - start_time, outpara
        )
    elif args.method == "pmprscs_auto":
        if not args.pm:
            parser.error("You must specify a Precision Matrix Folder Path.")
        if not args.snp:
            parser.error("You must specify a Snplists Folder Path.")
        if not args.sumstats:
            parser.error("You must specify a Sumstats File.")
        start_time = time.time()
        sumstats_list = pm.read.sumstats_read(args.sumstats)
        pm.filter.filter_sumstats(sumstats_list)
        pm.filter.filter_by_unique_sumstats(sumstats_list)
        PM = pm.read.PM_read(args.pm)
        snplist = pm.read.snplist_read(args.snp)
        pm.filter.filter_by_PM(PM, snplist)
        pm.filter.filter_by_unique_snplist(snplist)
        if not args.unnormal:
            pm.filter.normalize_PM_parallel(PM, para)
        sumstats = pm.filter.filter_by_sumstats_parallel(
            PM, snplist, sumstats_list, para
        )
        if args.N != None:
            pm.generate.generate_N(sumstats, args.N)
        pm.check.check_same_rsid(snplist, sumstats)
        beta_pmprscs_auto, outpara = pm.pmprscs.pmprscs_auto(
            PM, snplist, sumstats, para
        )
        end_time = time.time()
        pm.write.sumstats_beta_write(
            sumstats, beta_pmprscs_auto, args.out, end_time - start_time, outpara
        )
    elif args.method == "normalizepm":
        if not args.pm:
            parser.error("You must specify a Precision Matrix Folder Path.")
        PM = pm.Read.PM_read(args.pm)
        pm.Filter.normalize_PM_parallel(PM, para)
        pm.Write.PM_write(PM, args.out)
    else:
        parser.error(
            "You must choose --method in {pmldpred_auto, pmldpred_grid, pmprscs, normalizepm}"
        )


if __name__ == "__main__":
    main()



from pmpred.snp_read import read
from pmpred.snp_read import filter
from pmpred.snp_read import write
from pmpred.snp_read import check

from pmpred.model import predict
from pmpred.model import ldpred
from pmpred.model import pmldpred
from pmpred.model import prscs
from pmpred.model import pmprscs

from pmpred.simulation import generate



from scipy.stats import linregress
import numpy as np


def check_r_squared_phe(phestats, beta):
    y = phestats["Phenotype"].astype(float)
    x = phestats["X"] @ np.array(beta).astype(float)
    slope, intercept, r_value, p_value, std_err = linregress(x, y)
    return r_value**2


def check_r_squared_beta(beta1, beta2):
    slope, intercept, r_value, p_value, std_err = linregress(beta1, beta2)
    return r_value**2


def normalize(phestats):
    mean = phestats["X"].mean(axis=0)
    stddev = phestats["X"].std(axis=0)
    phestats["X"] = (phestats["X"] - mean) / stddev


def check_h2(PM, beta):
    h2 = 0
    for i in range(len(PM)):
        h2 += np.dot(beta[i], PM[i]["LD"].toarray() @ beta[i])
    return h2



# read precision matrix

import os
from scipy.sparse import csr_matrix


def snplist_read(snplist_folder_path):
    snplist = []
    for snplist_file in sorted(
        os.listdir(snplist_folder_path),
        key=lambda x: (
            int(x.split("_")[1][3:]),
            int(x.split("_")[2]),
            int(x.split("_")[3].split(".")[0]),
        ),
    ):
        snplist_block = {}
        with open(snplist_folder_path + "/" + snplist_file, "r") as file:
            tot = 0
            title_list = []
            index_dict = {}
            for line in file:
                line_list = line.strip().split(",")
                if tot > 0 and line_list[0] in index_dict:
                    continue
                if tot == 0:
                    title_list = line_list
                    for i in range(len(title_list)):
                        snplist_block[title_list[i]] = []
                else:
                    for i in range(len(title_list)):
                        snplist_block[title_list[i]].append(line_list[i])
                    index_dict[snplist_block["index"][tot - 1]] = tot - 1
                tot = tot + 1
        snplist_block["filename"] = snplist_file
        snplist_block["rsid"] = snplist_block.pop("site_ids")
        snplist_block["REF"] = snplist_block.pop("anc_alleles")
        snplist_block["ALT"] = snplist_block.pop("deriv_alleles")
        snplist.append(snplist_block)
        print("Read snplist file:", snplist_file)
    return snplist


def PM_read(precision_folder_path):
    PM = []
    for PM_file in sorted(
        os.listdir(precision_folder_path),
        key=lambda x: (
            int(x.split("_")[1][3:]),
            int(x.split("_")[2]),
            int(x.split("_")[3].split(".")[0]),
        ),
    ):
        PM_block = {}
        rows = []
        cols = []
        data = []
        with open(precision_folder_path + "/" + PM_file, "r") as file:
            for line in file:
                row_idx, col_idx, value = map(float, line.strip().split(","))
                rows.append(int(row_idx))
                cols.append(int(col_idx))
                data.append(value)
        PM_block["precision"] = csr_matrix((data, (rows, cols)))
        PM_block["filename"] = PM_file
        PM.append(PM_block)
        print("Read Precision matrix file:", PM_file)
    return PM


# def sumstats_read(sumstats_path):
#     sumstats = {}
#     with open(sumstats_path, "r") as file:
#         tot = 0
#         for line in file:
#             line_list = line.strip().split(",")
#             if tot == 0:
#                 title_list = line_list
#                 for i in range(len(title_list)):
#                     sumstats[title_list[i]] = []
#             else:
#                 for i in range(len(title_list)):
#                     sumstats[title_list[i]].append(line_list[i])
#             tot = tot + 1
#     sumstats["REF"] = sumstats.pop("a0")
#     sumstats["ALT"] = sumstats.pop("a1")
#     return sumstats


def sumstats_read(sumstats_path, split, head_name):
    sumstats = {}
    with open(sumstats_path, "r") as file:
        tot = 0
        for line in file:
            line_list = line.strip().split(split)
            if tot == 0:
                title_list = line_list
                for i in range(len(title_list)):
                    sumstats[title_list[i]] = []
            else:
                for i in range(len(title_list)):
                    sumstats[title_list[i]].append(line_list[i])
            tot = tot + 1
            if tot % 100000 == 0:
                print("Read sumstats line:", tot)
    for key in head_name.keys():
        if head_name[key] in sumstats:
            sumstats[key] = sumstats.pop(head_name[key])
    return sumstats


def vcf_read(vcf_path):
    vcfstats = {}
    with open(vcf_path, "r") as file:
        tot = 0
        for line in file:
            if line[0:2] == "##":
                continue
            line_list = line.strip().split("\t")
            if tot == 0:
                title_list = line_list
                for i in range(len(title_list)):
                    vcfstats[title_list[i]] = []
            else:
                for i in range(len(title_list)):
                    if title_list[i] == "ID" or title_list[i] == "REF":
                        vcfstats[title_list[i]].append(line_list[i])
                    elif len(line_list[i]) == 3 and line_list[i][1] == "/":
                        v = line_list[i]
                        if v[0] == ".":
                            a = 0
                        else:
                            a = int(v[0])
                        if v[2] == ".":
                            b = 0
                        else:
                            b = int(v[2])
                        vcfstats[title_list[i]].append(a + b)
            tot = tot + 1
            if tot % 1000 == 0:
                print("Read vcf file line:", tot)
    for key in title_list:
        if len(vcfstats[key]) == 0:
            vcfstats.pop(key)
    vcfstats["rsid"] = vcfstats.pop("ID")
    return vcfstats


def fam_read(fam_path):
    famstats = {}
    title_list = [
        "IndividualID",
        "Phenotype",
    ]
    for i in range(len(title_list)):
        famstats[title_list[i]] = []
    with open(fam_path, "r") as file:
        for line in file:
            line_list = line.strip().split(" ")
            famstats["IndividualID"].append(line_list[0] + "_" + line_list[1])
            famstats["Phenotype"].append(line_list[2])
    return famstats


def beta_read(beta_path):
    betastats = {"rsid": [], "REF": [], "beta": []}
    with open(beta_path, "r") as file:
        tot = 0
        for line in file:
            line_list = line.strip().split("\t")
            if tot > 0:
                betastats["rsid"].append(line_list[0])
                betastats["REF"].append(line_list[1])
                betastats["beta"].append(line_list[2])
            tot += 1
    return betastats



def check_same_rsid(snplist, sumstats):
    for i in range(len(snplist)):
        if snplist[i]["rsid"] != sumstats[i]["rsid"]:
            raise Exception("The rsids of the snplist are not same as sumstats!")


def check_unique_rsid_sumstats(sumstats):
    rsid_set = set()
    for i in range(len(sumstats)):
        rsid_block_set = set(sumstats[i]["rsid"])
        s = len(rsid_set)
        rsid_set.update(rsid_block_set)
        if s + len(sumstats[i]["rsid"]) != len(rsid_set):
            raise Exception("The rsids of sumstats are have same element!")



import numpy as np
from scipy.sparse.linalg import spsolve
from scipy.sparse import csr_matrix, eye


def get_marginal_beta(sumstats):
    beta_marginal = []
    for i in range(len(sumstats)):
        beta_marginal.append(list(map(float, sumstats[i]["beta"])))
    return beta_marginal


def ldpred_inf(PM, snplist, sumstats, para):
    beta_inf_set = []
    m = 0
    for i in range(len(sumstats)):
        m += len(sumstats[i]["N"])
    for i in range(len(PM)):
        if len(sumstats[i]["N"]) == 0:
            beta_inf_set.append([])
            continue
        N = np.array(sumstats[i]["N"]).astype(float)
        beta_se = np.array(sumstats[i]["beta_se"]).astype(float)
        beta = np.array(sumstats[i]["beta"]).astype(float)
        scale = np.sqrt(N * beta_se**2)
        # scale = 1
        beta_hat = beta / scale
        LD = PM[i]["LD"][snplist[i]["index"]][:, snplist[i]["index"]]
        LD += eye(LD.shape[0], format="csr") * (m / (para["h2"] * para["N"]))
        beta_inf = spsolve(LD, beta_hat)
        beta_inf_set.append(beta_inf * scale)
    return beta_inf_set, {"p": para["p"], "h2": para["h2"]}


def ldpred_gibbs_one_sampling(PM, snplist, beta_hat, N, M, para):
    LD = PM["LD"][snplist["index"]][:, snplist["index"]]
    # LD = inv(PM["precision"])
    if isinstance(LD[0], np.float64):
        LD = csr_matrix([[LD[0]]])
    m = len(beta_hat)
    h2 = para["h2"]
    p = para["p"]
    curr_beta = np.zeros(m)
    avg_beta = np.zeros(m)
    dotprods = np.zeros(m)
    h2_per_var = h2 / (M * p)
    inv_odd_p = (1 - p) / p
    for k in range(-para["burn_in"], para["num_iter"]):
        for j in range(m):
            res_beta_hat_j = beta_hat[j] - (dotprods[j] - curr_beta[j])
            # res_beta_hat_j = beta_hat[j] - dotprods[j]
            C1 = h2_per_var * N[j]
            C2 = 1 / (1 + 1 / C1)
            C3 = C2 * res_beta_hat_j
            C4 = C2 / N[j]
            post_p_j = 1 / (1 + inv_odd_p * np.sqrt(1 + C1) * np.exp(-C3 * C3 / C4 / 2))
            diff = -curr_beta[j]
            # if post_p_j < p:
            #     curr_beta[j] = 0
            if post_p_j > np.random.rand():
                curr_beta[j] = np.random.normal(C3, np.sqrt(C4))
                diff += curr_beta[j]
            else:
                curr_beta[j] = 0
            if k >= 0:
                avg_beta[j] += C3 * post_p_j
            if diff != 0:
                dotprods += LD[:, j].toarray().flatten() * diff
    return avg_beta / para["num_iter"]


def ldpred_grid(PM, snplist, sumstats, para):
    beta_gibbs_set = []
    M = 0
    for i in range(len(sumstats)):
        M += len(sumstats[i]["beta"])
    for i in range(len(PM)):
        if len(sumstats[i]["beta"]) == 0:
            beta_gibbs_set.append([])
            continue
        N = np.array(sumstats[i]["N"]).astype(float)
        beta_se = np.array(sumstats[i]["beta_se"]).astype(float)
        beta = np.array(sumstats[i]["beta"]).astype(float)
        # scale = np.sqrt(N * beta_se**2 + beta**2)
        scale = 1
        beta_hat = beta / scale
        beta_gibbs = ldpred_gibbs_one_sampling(PM[i], snplist[i], beta_hat, N, M, para)
        beta_gibbs_set.append(beta_gibbs * scale)
    return beta_gibbs_set, {"p": para["p"], "h2": para["h2"]}


def ldpred_auto(PM, snplist, sumstats, para):
    p = para["p"]
    h2 = para["h2"]
    curr_beta = []
    avg_beta = []
    dotprods = []
    M = 0
    for i in range(len(sumstats)):
        m = len(sumstats[i]["beta"])
        M += m
        curr_beta.append(np.zeros(m))
        avg_beta.append(np.zeros(m))
        dotprods.append(np.zeros(m))

    for k in range(-para["burn_in"], para["num_iter"]):
        Mc = 0
        print("step:", k, "p:", p, "h2:", h2)
        h2_per_var = h2 / (M * p)
        inv_odd_p = (1 - p) / p
        h2 = 0
        for i in range(len(PM)):
            if len(sumstats[i]["beta"]) == 0:
                continue
            N = np.array(sumstats[i]["N"]).astype(float)
            beta_hat = np.array(sumstats[i]["beta"]).astype(float)
            LD = PM[i]["LD"][snplist[i]["index"]][:, snplist[i]["index"]]
            if isinstance(LD[0], np.float64):
                LD = csr_matrix([[LD[0]]])
            m = len(beta_hat)
            for j in range(m):
                res_beta_hat_j = beta_hat[j] - (dotprods[i][j] - curr_beta[i][j])
                C1 = h2_per_var * N[j]
                C2 = 1 / (1 + 1 / C1)
                C3 = C2 * res_beta_hat_j
                C4 = C2 / N[j]
                post_p_j = 1 / (
                    1 + inv_odd_p * np.sqrt(1 + C1) * np.exp(-C3 * C3 / C4 / 2)
                )
                diff = -curr_beta[i][j]
                # if post_p_j < p:
                #     curr_beta[j] = 0
                if post_p_j > np.random.rand():
                    curr_beta[i][j] = np.random.normal(C3, np.sqrt(C4))
                    diff += curr_beta[i][j]
                    Mc += 1
                else:
                    curr_beta[i][j] = 0
                if k >= 0:
                    avg_beta[i][j] += C3 * post_p_j
                if diff != 0:
                    dotprods[i] += LD[:, j].toarray().flatten() * diff
            h2 += np.dot(curr_beta[i], LD.dot(curr_beta[i]))
        p = np.random.beta(1 + Mc, 1 + M - Mc)
    beta_auto_set = []
    for i in range(len(sumstats)):
        if len(sumstats[i]["beta"]) == 0:
            beta_auto_set.append(np.array([]))
        else:
            beta_auto_set.append(avg_beta[i] / para["num_iter"])
    return beta_auto_set, {"p": p, "h2": h2}



def sumstats_beta_write(sumstats, beta, output_path, run_time, outpara, head_name):
    # output_key = ["rsid", "CHR", "POS", "REF", "ALT", "beta", "beta_se", "P", "N"]
    output_key = ["rsid", "REF"]
    with open(output_path, "w") as f:
        f.write(f"## Running time: {run_time:.4f} seconds\n")
        f.write(f"##")
        for key in outpara.keys():
            f.write(f" {key} = {outpara[key]:.4f}")
        f.write("\n")
        f.write("\t".join([head_name[key] for key in output_key]))
        f.write("\tbeta_joint\n")
        for i in range(len(sumstats)):
            for j in range(len(beta[i])):
                for key in output_key:
                    if key in sumstats[i]:
                        if isinstance(sumstats[i][key][j], str):
                            f.write(sumstats[i][key][j] + "\t")
                        elif key == "N":
                            f.write(f"{int(sumstats[i][key][j])}\t")
                        else:
                            f.write(f"{sumstats[i][key][j]:.6f}\t")
                    else:
                        f.write("NA\t")
                f.write(f"{beta[i][j]:.6f}")
                f.write("\n")


def PM_write(PM, output_folder_path):
    for i in range(len(PM)):
        with open(output_folder_path + "/" + PM[i]["filename"], "w") as f:
            Q = PM[i]["precision"].tocoo()
            for (row, col), value in zip(zip(Q.row, Q.col), Q.data):
                f.write(f"{row},{col},{value:.6f}\n")



import numpy as np
from scipy.sparse.linalg import gmres
from joblib import Parallel, delayed


def sparse_LD_times(P, x, Pidn0, para):
    if len(Pidn0) == 0:
        return np.array([])
    y = np.zeros(P.shape[0])
    y[Pidn0] = x
    return gmres(P, y, rtol=para["rtol"])[0][Pidn0]


def pmpred_Q_times(P, x, n, Pid, sigma2, para):
    y = np.zeros(P.shape[0])
    y[Pid] = x * np.sqrt(n)
    return sigma2 * np.sqrt(n) * gmres(P, y, rtol=para["rtol"])[0][Pid]


def pmpred_grid_subprocess(subinput):
    (
        PM,
        snplist,
        beta_hat,
        curr_beta,
        R_curr_beta,
        N,
        h2_per_var,
        inv_odd_p,
        h2,
        para,
        i,
        k,
    ) = subinput
    res_beta_hat = beta_hat * np.sqrt(1 - h2) - (R_curr_beta - curr_beta)
    n = N / (1 - h2)
    C1 = h2_per_var * n
    C2 = 1 / (1 + 1 / C1)
    C3 = C2 * res_beta_hat
    C4 = C2 / n
    m = len(beta_hat)
    post_p = 1 / (1 + inv_odd_p * np.sqrt(1 + C1) * np.exp(-C3 * C3 / C4 / 2))
    z = (np.random.rand(m) < post_p).astype(int)
    id0 = np.where(z == 0)[0]
    idn0 = np.where(z != 0)[0]
    Pidn0 = snplist["index"][idn0]
    mean_beta = np.zeros(m)
    if len(Pidn0) != 0:
        P = PM["precision"].copy()
        P[Pidn0, Pidn0] += h2_per_var * n[idn0]
        delta = np.sqrt(1 - h2) * beta_hat[idn0] * n[idn0]
        mean = h2_per_var * (
            delta - pmpred_Q_times(P, delta, n[idn0], Pidn0, h2_per_var, para)
        )
        curr_beta[idn0] = np.random.randn(len(idn0))
        x = curr_beta[idn0].copy()
        l = 1
        while np.max(np.abs(x)) / np.max(np.abs(curr_beta)) > para["taylor_rtol"]:
            x = (
                -(0.5 - l + 1)
                / l
                * pmpred_Q_times(P, x, n[idn0], Pidn0, h2_per_var, para)
            )
            curr_beta[idn0] += x
            l += 1
        if i % 137 == 0:
            print("Run ldpred_auto block:", i, "Iteration:", k, "Taylor:", l)
        curr_beta[idn0] *= np.sqrt(h2_per_var)
        curr_beta[idn0] += mean
        mean_beta[idn0] = mean
    curr_beta[id0] = 0
    R_curr_beta = sparse_LD_times(PM["precision"], curr_beta, snplist["index"], para)
    return curr_beta, R_curr_beta, mean_beta


def pmldpred_grid(PM, snplist, sumstats, para):
    p = para["p"]
    h2 = para["h2"]
    M = 0
    beta_pmldpred_auto = []
    curr_beta = []
    avg_beta = []
    beta_hat = []
    N = []
    scale_size = []
    R_curr_beta = []
    for i in range(len(PM)):
        M += len(sumstats[i]["beta"])
    for i in range(len(PM)):
        beta_hat.append(np.array(sumstats[i]["beta"]).astype(float))
        N.append(np.array(sumstats[i]["N"]).astype(float))
        scale_size.append(
            np.sqrt(N[i]) * np.array(sumstats[i]["beta_se"]).astype(float)
        )
        beta_hat[i] = beta_hat[i] / scale_size[i]
        m = len(beta_hat[i])
        curr_beta.append(np.zeros(m))
        R_curr_beta.append(np.zeros(m))
        avg_beta.append(np.zeros(m))
        snplist[i]["index"] = np.array(snplist[i]["index"])
    for k in range(-para["burn_in"], para["num_iter"]):
        print("Run pmldpred_auto step:", k, "p:", p, "h2:", h2)
        h2_per_var = h2 / (M * p)
        inv_odd_p = (1 - p) / p
        subinput = []
        for i in range(len(PM)):
            subinput.append(
                (
                    PM[i],
                    snplist[i],
                    beta_hat[i],
                    curr_beta[i],
                    R_curr_beta[i],
                    N[i],
                    h2_per_var,
                    inv_odd_p,
                    h2,
                    para,
                    i,
                    k,
                )
            )
        results = Parallel(n_jobs=para["n_jobs"])(
            delayed(pmpred_grid_subprocess)(d) for d in subinput
        )
        for i in range(len(PM)):
            curr_beta[i], R_curr_beta[i], mean_beta = results[i]
            if k >= 0:
                avg_beta[i] += mean_beta
    for i in range(len(PM)):
        beta_pmldpred_auto.append(avg_beta[i] / para["num_iter"])
        beta_pmldpred_auto[i] = beta_pmldpred_auto[i] * scale_size[i]
        snplist[i]["index"] = snplist[i]["index"].tolist()
    outpara = {
        "p": p,
        "h2": h2,
        "burn_in": para["burn_in"],
        "num_iter": para["num_iter"],
        "log10rtol": np.log(para["rtol"]) / np.log(10),
        "log10taylor_rtol": np.log(para["taylor_rtol"]) / np.log(10),
        "prop": para["prop"],
    }
    return beta_pmldpred_auto, outpara


def pmpred_auto_subprocess(subinput):
    (
        PM,
        snplist,
        beta_hat,
        curr_beta,
        R_curr_beta,
        N,
        h2_per_var,
        inv_odd_p,
        h2,
        para,
        i,
        k,
    ) = subinput
    res_beta_hat = beta_hat * np.sqrt(1 - h2) - (R_curr_beta - curr_beta)
    # res_beta_hat = np.sqrt(1 - h2) * (beta_hat - R_curr_beta) + curr_beta
    # res_beta_hat = beta_hat - (R_curr_beta - curr_beta)
    n = N / (1 - h2)
    # n = N
    C1 = h2_per_var * n
    C2 = 1 / (1 + 1 / C1)
    C3 = C2 * res_beta_hat
    C4 = C2 / n
    m = len(beta_hat)
    post_p = 1 / (1 + inv_odd_p * np.sqrt(1 + C1) * np.exp(-C3 * C3 / C4 / 2))
    z = (np.random.rand(m) < post_p).astype(int)
    id0 = np.where(z == 0)[0]
    idn0 = np.where(z != 0)[0]
    Pidn0 = snplist["index"][idn0]
    mean_beta = np.zeros(m)
    if len(Pidn0) != 0:
        P = PM["precision"].copy()
        P[Pidn0, Pidn0] += h2_per_var * n[idn0]
        delta = np.sqrt(1 - h2) * beta_hat[idn0] * n[idn0]
        # delta = (
        #     np.sqrt(1 - h2) * beta_hat[idn0] + (1 - np.sqrt(1 - h2)) * R_curr_beta[idn0]
        # ) * n[idn0]
        # delta = beta_hat[idn0] * n[idn0]
        mean = h2_per_var * (
            delta - pmpred_Q_times(P, delta, n[idn0], Pidn0, h2_per_var, para)
        )
        # mean = np.linalg.solve(
        #     PM["LD"][Pidn0][:, Pidn0].toarray() + np.diag(1 / (h2_per_var * n[idn0])),
        #     beta_hat[idn0],
        # )
        # alpha = np.zeros(P.shape[0])
        # alpha[Pidn0] = np.sqrt(1 - h2) * beta_hat[idn0] * np.sqrt(n[idn0])
        # mean = (
        #     h2_per_var
        #     * np.sqrt(n[idn0])
        #     * gmres(P, PM["precision"].dot(alpha), rtol=para["rtol"])[0][Pidn0]
        # )
        curr_beta[idn0] = np.random.randn(len(idn0))
        x = curr_beta[idn0].copy()
        l = 1
        while np.max(np.abs(x)) / np.max(np.abs(curr_beta)) > para["taylor_rtol"]:
            x = (
                -(0.5 - l + 1)
                / l
                * pmpred_Q_times(P, x, n[idn0], Pidn0, h2_per_var, para)
            )
            curr_beta[idn0] += x
            l += 1
        if i % 137 == 0:
            print("Run ldpred_auto block:", i, "Iteration:", k, "Taylor:", l)
        curr_beta[idn0] *= np.sqrt(h2_per_var)
        curr_beta[idn0] += mean
        mean_beta[idn0] = mean
    curr_beta[id0] = 0
    R_curr_beta = sparse_LD_times(PM["precision"], curr_beta, snplist["index"], para)
    return curr_beta, R_curr_beta, len(idn0), np.dot(curr_beta, R_curr_beta), mean_beta


def pmldpred_auto(PM, snplist, sumstats, para):
    p = para["p"]
    h2 = para["h2"]
    M = 0
    beta_pmldpred_auto = []
    curr_beta = []
    avg_beta = []
    beta_hat = []
    N = []
    scale_size = []
    R_curr_beta = []
    for i in range(len(PM)):
        M += len(sumstats[i]["beta"])
    for i in range(len(PM)):
        beta_hat.append(np.array(sumstats[i]["beta"]).astype(float))
        N.append(np.array(sumstats[i]["N"]).astype(float))
        scale_size.append(
            np.sqrt(N[i]) * np.array(sumstats[i]["beta_se"]).astype(float)
        )
        beta_hat[i] = beta_hat[i] / scale_size[i]
        m = len(beta_hat[i])
        curr_beta.append(np.zeros(m))
        R_curr_beta.append(np.zeros(m))
        avg_beta.append(np.zeros(m))
        snplist[i]["index"] = np.array(snplist[i]["index"])
    h2_last = h2
    for k in range(-para["burn_in"], para["num_iter"]):
        Mc = 0
        h2 = min(h2, 1)
        h2 = (1 - para["prop"]) * h2_last + para["prop"] * h2
        h2_last = h2
        print("Run pmldpred_auto step:", k, "p:", p, "h2:", h2)
        h2_per_var = h2 / (M * p)
        inv_odd_p = (1 - p) / p
        subinput = []
        for i in range(len(PM)):
            subinput.append(
                (
                    PM[i],
                    snplist[i],
                    beta_hat[i],
                    curr_beta[i],
                    R_curr_beta[i],
                    N[i],
                    h2_per_var,
                    inv_odd_p,
                    h2,
                    para,
                    i,
                    k,
                )
            )
        results = Parallel(n_jobs=para["n_jobs"])(
            delayed(pmpred_auto_subprocess)(d) for d in subinput
        )
        h2 = 0
        for i in range(len(PM)):
            curr_beta[i], R_curr_beta[i], Mc_add, h2_add, mean_beta = results[i]
            Mc += Mc_add
            h2 += h2_add
            if k >= 0:
                avg_beta[i] += mean_beta
        p = np.random.beta(1 + Mc, 1 + M - Mc)
    for i in range(len(PM)):
        beta_pmldpred_auto.append(avg_beta[i] / para["num_iter"])
        beta_pmldpred_auto[i] = beta_pmldpred_auto[i] * scale_size[i]
        snplist[i]["index"] = snplist[i]["index"].tolist()
    outpara = {
        "p": p,
        "h2": h2,
        "burn_in": para["burn_in"],
        "num_iter": para["num_iter"],
        "log10rtol": np.log(para["rtol"]) / np.log(10),
        "log10taylor_rtol": np.log(para["taylor_rtol"]) / np.log(10),
        "prop": para["prop"],
    }
    return beta_pmldpred_auto, outpara



import numpy as np
from scipy.sparse import csr_matrix
from scipy.stats import norm


def generate_sumstats_beta(PM, snplist, para):
    sumstats = []
    beta_true = []
    m = 0
    for i in range(len(snplist)):
        m += len(snplist[i]["rsid"])
    si = 0
    sigma = para["h2"] / (m * para["p"])
    for i in range(len(snplist)):
        sumstats_block = {}
        mm = para["sumstats_block_size"]
        sumstats_block["N"] = np.ones(mm) * para["N"]
        sumstats_block["beta_se"] = (
            np.ones(mm) * np.sqrt(1 - para["h2"]) / np.sqrt(para["N"])
        )
        sumstats_block["rsid"] = np.array(snplist[i]["rsid"])[0:mm].tolist()
        si += para["PM_size"]
        beta_true.append([])
        if mm == 0:
            sumstats_block["beta"] = []
            sumstats.append(sumstats_block)
            continue
        for j in range(mm):
            if np.random.rand() < para["p"]:
                beta_true[i].append(np.random.normal(0, np.sqrt(sigma)))
            else:
                beta_true[i].append(0)
        beta_true[i] = np.array(beta_true[i])
        R = PM[i]["LD"].toarray()
        sumstats_block["beta"] = np.random.multivariate_normal(
            R @ beta_true[i], R * (1 - para["h2"]) / para["N"]
        )
        sumstats.append(sumstats_block)

    return sumstats, beta_true


def generate_phenotype(phestats, beta, para):
    phestats["Phenotype"] = phestats["X"] @ beta + np.random.randn(
        len(phestats["Phenotype"])
    ) * np.sqrt(1 - para["h2"])


def generate_PM_snplist(para):
    snplist = []
    PM = []
    si = 0
    for i in range(para["block_num"]):
        snplist_block = {}
        PM_block = {}
        N = para["PM_size"]
        snplist_block["rsid"] = [
            "rs{}".format(i) for i in range(si, si + para["sumstats_block_size"])
        ]
        si += N
        snplist_block["index"] = np.arange(para["sumstats_block_size"])
        M = np.random.rand(N, N) - 0.5
        M += M.T
        R = np.dot(M, M.T)
        D = np.sqrt(np.diag(R))
        PM_block["LD"] = R / np.outer(D, D)
        PM_block["precision"] = csr_matrix(np.linalg.inv(PM_block["LD"]))
        PM_block["LD"] = csr_matrix(PM_block["LD"])[0 : para["sumstats_block_size"]][
            :, 0 : para["sumstats_block_size"]
        ]
        snplist.append(snplist_block)
        PM.append(PM_block)
    return PM, snplist


def generate_beta_and_se_from_p_and_z(sumstats):
    print("Generate beta and beta_se from p value and z scores in sumstats")
    for i in range(len(sumstats)):
        sumstats[i]["beta"] = (
            np.abs(norm.ppf(np.array(sumstats[i]["p"]).astype(float) / 2))
            * np.sign(np.array(sumstats[i]["z"]).astype(float))
        ).tolist()
        sumstats[i]["beta_se"] = np.ones(len(sumstats[i]["beta"])).tolist()


def generate_N_in_sumstats_list(sumstats_list, N):
    print("Generate N in sumstats_list")
    sumstats_list["N"] = (np.ones(len(sumstats_list["rsid"])) * N).tolist()


def get_para():
    para = {}
    para["h2"] = 0.5
    para["p"] = 0.18
    para["sumstats_block_size"] = 10
    para["PM_size"] = 11
    para["block_num"] = 10
    para["burn_in"] = 50
    para["num_iter"] = 100
    para["N"] = 100000
    para["prop"] = 1
    para["taylor_rtol"] = 0.01
    para["h2_min"] = 0
    para["h2_max"] = 1
    para["p_min"] = 0
    para["p_max"] = 1
    para["prop"] = 0.1
    para["n_jobs"] = -1
    para["rtol"] = 1e-10
    para["prscs_a"] = 1.5
    para["prscs_b"] = 0.5
    return para



import numpy as np
import scipy.sparse as sp
from joblib import Parallel, delayed


def filter_by_PM(PM, snplist):
    for i in range(len(PM)):
        PMid = np.where(np.diff(PM[i]["precision"].indptr) != 0)[0]
        PM[i]["precision"] = PM[i]["precision"][PMid][:, PMid]
        for key in list(snplist[i].keys()):
            if isinstance(snplist[i][key], list):
                snplist[i][key] = np.array(snplist[i][key])[PMid].tolist()
        if i % 137 == 0:
            print("Filter_by_PM block:", i)
        snplist[i]["index"] = np.arange(len(snplist[i]["rsid"])).tolist()


def filter_sumstats(sumstats):
    base_map = {"A": "T", "T": "A", "C": "G", "G": "C"}
    N_mean = np.sum(np.array(sumstats["N"]).astype(int)) / len(sumstats["N"])
    i = 0
    while i < len(sumstats["N"]):
        if (
            int(sumstats["N"][i]) < N_mean / 3
            # and sumstats["REF"][i] in base_map.keys()
            # and sumstats["ALT"][i] in base_map.keys()
            and base_map[sumstats["REF"][i]] == sumstats["ALT"][i]
        ):
            for key in sumstats.keys():
                del sumstats[key][i]
        else:
            i += 1
        if i % 100000 == 0:
            print("Filter sumstats line:", i)


def normalize_PM_parallel_subprocess(subinput):
    P, i = subinput
    print("normalize_PM_subprocess block:", i)
    Pid = sorted(set(P.tocoo().row))
    D = np.zeros(P.shape[0])
    D[Pid] = np.sqrt(
        sp.linalg.spsolve(
            P[Pid][:, Pid].tocsc(), sp.eye(len(Pid), format="csc")
        ).diagonal()
    )
    return P.multiply(np.outer(D, D)).tocsr()


def normalize_PM_parallel(PM, para):
    subinput = []
    for i in range(len(PM)):
        subinput.append((PM[i]["precision"], i))
    results = Parallel(n_jobs=para["n_jobs"])(
        delayed(normalize_PM_parallel_subprocess)(d) for d in subinput
    )
    for i in range(len(PM)):
        PM[i]["precision"] = results[i]


def normalize_dense_matrix(A):
    diag_elements = np.diag(A)
    sqrt_diag_outer = np.sqrt(np.outer(diag_elements, diag_elements))
    return A / sqrt_diag_outer


def filter_by_sumstats_subprocess(subinput):
    rsid_sumstats, snplist_rsid, i = subinput
    rsid1 = [
        index for index, value in enumerate(snplist_rsid) if value in rsid_sumstats
    ]
    rsid2 = [
        rsid_sumstats[value]
        for index, value in enumerate(snplist_rsid)
        if value in rsid_sumstats
    ]
    if i % 137 == 0:
        print("Filter_by_sumstats parallel block:", i)
    return rsid1, rsid2


def filter_by_sumstats_parallel(PM, snplist, sumstats, para):
    sumstats_set = []
    rsid_sumstats = {value: index for index, value in enumerate(sumstats["rsid"])}
    subinput = []
    for i in range(len(PM)):
        subinput.append((rsid_sumstats, snplist[i]["rsid"], i))
    for key in list(sumstats.keys()):
        if isinstance(sumstats[key], list):
            sumstats[key] = np.array(sumstats[key])
    results = Parallel(n_jobs=para["n_jobs"], backend="threading")(
        delayed(filter_by_sumstats_subprocess)(d) for d in subinput
    )
    for i in range(len(PM)):
        rsid1, rsid2 = results[i]
        for key in list(snplist[i].keys()):
            if isinstance(snplist[i][key], list):
                snplist[i][key] = np.array(snplist[i][key])[rsid1].tolist()
        sumstats_block = {}
        for key in list(sumstats.keys()):
            if isinstance(sumstats[key], np.ndarray):
                sumstats_block[key] = sumstats[key][rsid2].tolist()
        sumstats_set.append(sumstats_block)
        if i % 137 == 0:
            print("Filter_by_sumstats results block:", i)
    return sumstats_set


def filter_by_unique_sumstats(sumstats):
    print("Make sumstats rsid unique")
    unique_indices = set()
    unique_indices = [
        unique_indices.add(x) or i
        for i, x in enumerate(sumstats["rsid"])
        if x not in unique_indices
    ]
    for key in sumstats.keys():
        if isinstance(sumstats[key], list):
            sumstats[key] = np.array(sumstats[key])[unique_indices].tolist()


def filter_by_unique_snplist(snplist):
    unique_indices = set()
    for i in range(len(snplist)):
        unique_block_indices = [
            unique_indices.add(x) or i
            for i, x in enumerate(snplist[i]["rsid"])
            if x not in unique_indices
        ]
        for key in snplist[i].keys():
            if isinstance(snplist[i][key], list):
                snplist[i][key] = np.array(snplist[i][key])[
                    unique_block_indices
                ].tolist()
        if i % 137 == 0:
            print("Make snplist unique block:", i)


def filter_by_sumstats(PM, snplist, sumstats):
    sumstats_set = []
    rsid_sumstats = {value: index for index, value in enumerate(sumstats["rsid"])}
    for i in range(len(PM)):
        rsid = [
            index
            for index, value in enumerate(snplist[i]["rsid"])
            if value in rsid_sumstats
        ]
        for key in list(snplist[i].keys()):
            if isinstance(snplist[i][key], list):
                snplist[i][key] = np.array(snplist[i][key])[rsid].tolist()
        rsid = [
            rsid_sumstats[value]
            for index, value in enumerate(snplist[i]["rsid"])
            if value in rsid_sumstats
        ]
        sumstats_block = {}
        for key in list(sumstats.keys()):
            if isinstance(sumstats[key], list):
                sumstats_block[key] = np.array(sumstats[key])[rsid].tolist()
        sumstats_set.append(sumstats_block)
        print("Filter_by_sumstats block:", i)
    return sumstats_set


def merge_vcf_fam(vcfstats, famstats):
    phestats = {}
    N = len(famstats["IndividualID"])
    M = len(vcfstats["rsid"])
    X = np.zeros((N, M))
    for i in range(N):
        if i % 1000 == 0:
            print("merge_vcf_fam Individual:", i)
        for j in range(M):
            X[i][j] = vcfstats[famstats["IndividualID"][i]][j]
    phestats["X"] = X
    phestats["rsid"] = vcfstats["rsid"][:]
    phestats["REF"] = vcfstats["REF"][:]
    # phestats["ALT"] = vcfstats["ALT"][:]
    phestats["Phenotype"] = np.array(famstats["Phenotype"])
    del vcfstats
    return phestats


def filter_by_vcf(betastats, phestats):
    rsid_phestats = {value: index for index, value in enumerate(phestats["rsid"])}
    rsid = [
        index for index, value in enumerate(betastats["rsid"]) if value in rsid_phestats
    ]
    for key in list(betastats.keys()):
        if isinstance(betastats[key], list):
            betastats[key] = np.array(betastats[key])[rsid].tolist()
    rsid_betastats = {value: index for index, value in enumerate(betastats["rsid"])}
    rsid = [
        index for index, value in enumerate(phestats["rsid"]) if value in rsid_betastats
    ]
    for key in list(phestats.keys()):
        if isinstance(phestats[key], list):
            phestats[key] = np.array(phestats[key])[rsid].tolist()
    phestats["X"] = phestats["X"][:, rsid]


def merge_beta(beta):
    beta_total = []
    for i in range(len(beta)):
        if isinstance(beta[i], list) == 0:
            beta_total += beta[i].tolist()
        else:
            beta_total += beta[i]
    return np.array(beta_total)


def filter_by_REF_ALT(betastats, phestats):
    for j in range(len(phestats["rsid"])):
        if betastats["REF"][j] != phestats["REF"][j]:
            phestats["X"][:, j] = 2 - phestats["X"][:, j]


# def PM_get_LD(PM):
#     for i in range(len(PM)):
#         PM[i]["LD"] = inv(PM[i]["precision"])



import numpy as np
from scipy.stats import geninvgauss
from joblib import Parallel, delayed


def prscs_auto_subprocess_beta(subinput):
    (D, sigma2, N, psi, beta_hat) = subinput
    D = D.toarray()
    cov = sigma2 / N * np.linalg.inv(D + np.diag(1 / psi))
    mu = np.linalg.solve(D + np.diag(1 / psi), beta_hat)
    curr_beta = np.random.multivariate_normal(mu, cov)
    return curr_beta


def prscs_auto_subprocess_sigma2(subinput):
    curr_beta, beta_hat, D, psi = subinput
    sigma2_para = -2 * np.dot(curr_beta, beta_hat) + np.dot(
        curr_beta,
        D.toarray() @ curr_beta + curr_beta / psi,
    )
    sum_beta = np.dot(curr_beta, curr_beta / psi)
    return sigma2_para, sum_beta


def prscs_auto_subprocess_psi(subinput):
    N, sigma2, curr_beta, a, delta = subinput
    khi = N / sigma2 * curr_beta**2
    psi = geninvgauss.rvs(a - 1 / 2, np.sqrt(2 * delta * khi)) * np.sqrt(
        khi / (2 * delta)
    )
    psi[psi > 1] = 1
    return psi


def prscs_auto_subprocess_delta(subinput):
    a, b, psi, phi = subinput
    return np.random.gamma(a + b, 1 / (psi + phi))


def prscs_auto(PM, snplist, sumstats, para):
    M = 0
    a = para["prscs_a"]
    b = para["prscs_b"]
    curr_beta = []
    avg_beta = []
    beta_hat = []
    N = []
    N_total = 0
    scale_size = []
    delta = []
    psi = []
    phi = 1
    sigma2 = 1
    beta_prscs = []
    for i in range(len(PM)):
        M += len(sumstats[i]["beta"])
    for i in range(len(PM)):
        beta_hat.append(np.array(sumstats[i]["beta"]).astype(float))
        N.append(np.array(sumstats[i]["N"]).astype(float))
        N_total += np.sum(N[i])
        scale_size.append(
            np.sqrt(N[i] * np.array(sumstats[i]["beta_se"]).astype(float) ** 2)
        )
        beta_hat[i] = beta_hat[i] / scale_size[i]
        m = len(beta_hat[i])
        curr_beta.append(np.zeros(m))
        delta.append(np.random.gamma(b, 1 / phi, size=m))
        psi.append(np.random.gamma(a, 1 / delta[i]))
        avg_beta.append(np.zeros(m))
        snplist[i]["index"] = np.array(snplist[i]["index"])
    N_total = N_total / M
    for k in range(-para["burn_in"], para["num_iter"]):
        subinput = []
        for i in range(len(PM)):
            subinput.append(
                (
                    PM[i]["LD"][snplist[i]["index"]][:, snplist[i]["index"]],
                    sigma2,
                    N[i],
                    psi[i],
                    beta_hat[i],
                )
            )
        curr_beta = Parallel(n_jobs=para["n_jobs"])(
            delayed(prscs_auto_subprocess_beta)(d) for d in subinput
        )
        subinput = []
        for i in range(len(PM)):
            subinput.append(
                (
                    curr_beta[i],
                    beta_hat[i],
                    PM[i]["LD"][snplist[i]["index"]][:, snplist[i]["index"]],
                    psi[i],
                )
            )
        results = Parallel(n_jobs=para["n_jobs"])(
            delayed(prscs_auto_subprocess_sigma2)(d) for d in subinput
        )
        sigma2_para = 0
        sum_beta = 0
        for i in range(len(PM)):
            sigma2_para += results[i][0]
            sum_beta += results[i][1]
        err = max(N_total / 2 * (1 + sigma2_para), N_total / 2 * sum_beta)
        sigma2 = 1 / np.random.gamma((N_total + M) / 2, 1 / err)
        subinput = []
        for i in range(len(PM)):
            subinput.append((N[i], sigma2, curr_beta[i], a, delta[i]))
        psi = Parallel(n_jobs=para["n_jobs"])(
            delayed(prscs_auto_subprocess_psi)(d) for d in subinput
        )
        subinput = []
        for i in range(len(PM)):
            subinput.append((a, b, psi[i], phi))
        delta = Parallel(n_jobs=para["n_jobs"])(
            delayed(prscs_auto_subprocess_delta)(d) for d in subinput
        )
        sum_delta = 0
        for i in range(len(PM)):
            sum_delta += np.sum(delta[i])
        w = np.random.gamma(1, 1 / (1 + phi))
        phi = np.random.gamma(M * b + 1 / 2, 1 / (sum_delta + w))
        if k >= 0:
            for i in range(len(PM)):
                avg_beta[i] += curr_beta[i]
        print("Run prscs_auto step:", k, "phi:", phi, "w:", w)
    for i in range(len(PM)):
        beta_prscs.append(avg_beta[i] / para["num_iter"])
    return beta_prscs, {"phi": phi, "w": w}



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="PMpred",
    version="1.0.4",
    url="https://github.com/WiuYuan/pmpred",
    author="Wen Yuan",
    author_email="wiuwenyuan@gmail.com",
    description="A Python package that adjusts GWAS summary statistics for the effects of Sparse Precision Matrix (PM)",
    keywords="Polygenic Risk Scores, GWAS, Linkage Disequilibrium, Risk Prediction, Precision Matrix",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    python_requires=">=3.6",
    install_requires=["numpy", "scipy", "joblib"],
    entry_points={
        "console_scripts": [
            "pmpred=pmpred.main:main",
        ],
    },
)
